//#########################################################################
// spihyper.s
//
// Write/read a buffer with SPI Modul to from MMC/SD/SDHC.
// Very fast operation.
// SPI speed has to be F_CPU/2 for these routines !
//
// New version for ATMega644 which does not support "sbis SPSR, SPIF".
// Uses "in" and "sbrs" instead.
//
//#########################################################################
// Last change: 25.05.2008
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################

#define __ASSEMBLER__ 1
#define __SFR_OFFSET 0

#include <avr/io.h>

#include "mmc_spi.h"

#ifdef HYPER_SPI_WRITE
/** \ingroup mmc_spi
    \fn void write_spi_hyper(size_t len, const void *src)
    \brief Copy a memory area to SPI using fastest method.
           SPI speed has to be F_CPU/2 for this routine !
           
__tmp_reg__ 	= r0
len		= r24,r25
*src		= r22,r23
    \returns Nothing  */

	.text
	.global	write_spi_hyper
	.type	write_spi_hyper, @function
write_spi_hyper:
        push r30
        push r31
        
	mov	r30, r22                ; copy adress of buffer to Z
	mov	r31, r23


write_hyper_loop:
	ld	r22, Z+                  ;2
	out	SPDR, r22                ;1

	; Don't check SPIF. Simply do a loop
	; with 16+1 CPU cycles til next SPDR write

	sbiw	r24, 0          	;2 two nops
	sbiw	r24, 0          	;2 two nops
	sbiw	r24, 0          	;2 two nops
	sbiw	r24, 0          	;2 two nops
	sbiw	r24, 0          	;2 two nops
        nop    				;1

	sbiw	r24, 1          	;2 decrement len for first byte sent
	brne write_hyper_loop           ;2 if branch, 1 if not

        ldi	r23, 0xFF		; Dummy byte for spi write
        				; does not work if dummy byte is not 0xFF
wl1:
;       	sbis	SPSR, SPIF		;1 we have to read SPSR here to clear SPIF with next SPDR read/write !
	in r0,SPSR			; on ATMega644 sbis does not work !
	sbrs r0,SPIF
       	rjmp	wl1                     ;2 if not skipped

	out	SPDR, r23               ;1 read and ignore CRC16 part1
wl2:
;       	sbis	SPSR, SPIF		;1
	in r0,SPSR			
	sbrs r0,SPIF
       	rjmp	wl2                     ;2 if not skipped

	out	SPDR, r23               ;1 read and ignore CRC16 part2
        pop r31                         ; 2
        pop r30                         ; 2

wl3:
;       	sbis	SPSR, SPIF		;1
	in r0,SPSR			
	sbrs r0,SPIF
       	rjmp	wl3                     ;2 if not skipped

	ret                             ; 5
#endif //HYPER_SPI_WRITE

#ifdef HYPER_SPI_READ
/** \ingroup mmc_spi
    \fn void read_spi_hyper(size_t len,const void *dest)
    \brief Copy from SPI to a memory area.
           SPI speed has to be F_CPU/2 for this routine !
           
__tmp_reg__ 	= r0
len		= r24,r25
*dest		= r22,r23
    \returns Nothing  */
    
	.text
	.global	read_spi_hyper
	.type	read_spi_hyper, @function
read_spi_hyper:
        push r30
        push r31

	mov	r30, r22                ; copy adress of buffer to Z
	mov	r31, r23

        ldi	r23, 0xFF		; Dummy byte for spi write
        				; does not work if dummy byte is not 0xFF
	out	SPDR, r23               ; start shift in first byte 
	
rl1:
;       	sbis	SPSR, SPIF
	in r0,SPSR			
	sbrs r0,SPIF
       	rjmp	rl1

      	in	r22, SPDR 		; get first byte, but store later
	out	SPDR, r23               ; start shift in next byte
	                                ; we have to do 16+1 CPU cycles to next SPDR write
	                                 
	sbiw	r24, 1          	;2 decrement len for first byte read
	sbiw	r24, 0          	;2 two nops
	sbiw	r24, 0          	;2 two nops

read_hyper_loop:
       	st	Z+, r22			;2 store byte to buffer
	sbiw	r24, 1          	;2 decrement len
	sbiw	r24, 0          	;2 two nops
	sbiw	r24, 0          	;2 two nops
	sbiw	r24, 0          	;2 two nops

       	in	r22, SPDR               ;1 read last received byte
      	out	SPDR, r23               ;1 start shift in next byte
	sbiw	r24, 0          	;2 two nops

	sbiw	r24, 0          	;2 compare r24,r25 to zero
	brne read_hyper_loop            ;2 if branch, 1 if not

; last SPDR write is shifting in crc part1 at this point
       	st	Z, r22 			; store last received byte
rl2:
;       	sbis	SPSR, SPIF
	in r0,SPSR			
	sbrs r0,SPIF
       	rjmp	rl2

; shift in crc part2
       	out	SPDR, r23
        pop r31                         ; 2
        pop r30                         ; 2
rl3:
;       	sbis	SPSR, SPIF
	in r0,SPSR			
	sbrs r0,SPIF
       	rjmp	rl3

	ret                             ; 5
#endif //HYPER_SPI_READ

